/*  

    YABASIC ---  a simple Basic Interpreter
    written by Marc-Oliver Ihm 1995-99
    homepage: www.yabasic.de
    
    symbol.c --- code for control structures, symbol and stack management
    
    This file is part of yabasic and may be copied only 
    under the terms of either the Artistic License or 
    the GNU General Public License (GPL), both of which 
    can be found at www.yabasic.de

*/

/* ------------- includes ---------------- */

#ifndef YABASIC_INCLUDED
#include "yabasic.h"       /* all prototypes and structures */
#endif


/* ------------- external references ---------------- */

extern int yylineno;   /* current line number */
extern int yyparse();  /* call bison parser */


/* ------------- local functions ---------------- */

static struct symbol *create_symbol(int,char *); /* create a new symbol */
static void link_to_sym(struct command *,struct symbol *); /* link command into list associated with symbol */
static void unlink_from_sym(struct symbol *); /* unlink all commands from symbol */
static void reshufflestack(struct stackentry *); /* reorganize stack for function call */
static void link_label(struct command *); /* link label into list of labels */
static int count_args(int); /* count number of arguments on stack */
static struct command *search_label(char *,int,int *); /* search label */
static void stackdesc(int,char *); /* give back string describing stackentry */
static void freesym(struct symbol *); /* free contents of symbol */
static int ind_to_off(int *,int *); /* convert array of indices to single offset */
static void off_to_ind(int,int *,int *); /* convert a single offset to an array of indices */


/* ------------- global variables ---------------- */

static struct symstack *symroot=NULL; /* first element in symbol list */
static struct symstack *symhead=NULL; /* last element ind symbol list */
struct stackentry *stackroot; /* lowest element in stack */
struct stackentry *stackhead; /* topmost element in stack */
static int labelcount=0; /* count self-generated labels */
static struct command *labelroot=NULL; /* first label among commands */
static struct command *labelhead=NULL; /* last label seen so far */
extern char *current_function; /* name of currently defined function */

/* ------------- subroutines ---------------- */


void pushsymlist(void) /* push a new list of symbols on symbol stack */
{
  struct symstack *new;

  new=my_malloc(sizeof(struct symstack));
  if (symhead)
    symhead->next_in_stack=new;
  else
    symroot=new; /* first time called */
  new->prev_in_stack=symhead;
  new->next_in_stack=NULL;
  new->next_in_list=NULL;
  symhead=new;
}

  
void popsymlist(void) /* pop list of symbols and free symbol contents */
{
  struct symstack *prevstack;
  struct symbol *currsym,*nextsym;
  
  currsym=symhead->next_in_list;
  while(currsym) { /* loop through symbol list */
    unlink_from_sym(currsym);
    freesym(currsym);
    nextsym=currsym->next_in_list;
    free(currsym);
    currsym=nextsym;
  }
  if (infolevel>=DIAGNOSTIC) error(DIAGNOSTIC,"removing symbol list");
  prevstack=symhead->prev_in_stack;
  free(symhead);
  prevstack->next_in_stack=NULL;
  symhead=prevstack;
}


static void freesym(struct symbol *s) /* free contents of symbol */
{
  int i;
  int total;

  struct array *ar;
  if (s->link) return; /* it's a link, don't remove memory */
  if (s->type==sySTRING) {
    free(s->pointer);
  } else if (s->type==syARRAY) {
    if (infolevel>=DIAGNOSTIC) {
      sprintf(string,"removing array '%s()'",s->name);
      error(DIAGNOSTIC,string);
    }
    ar=s->pointer;
    if (ar->dimension>0) {
      /* count total amount of memory */
      total=1;
      for(i=0;i<ar->dimension;i++) total*=(ar->bounds)[i];
      if (ar->type=='s') { /* string array */
	for(i=0;i<total;i++) free(*((char **)ar->pointer+i));
      }
      free(ar->pointer);
    }
    free(ar);
  }
  return;
}


struct symbol *get_sym(char *name,struct command *cmd,int type,int add) 
/* get the value of a symbol, or create it with given type */
{
  struct symstack *currstack;
  struct symbol **currsym;
  struct symbol *prelink;
  struct symbol *new;
  int stackcount=0;
  int symbolcount=0;
  int linked=FALSE;


  if (!name) return NULL;
  /* go through all lists */
  currstack=symhead;
  while(currstack) {   /* search 'til last element of stack */
    if (currstack->prev_in_stack==NULL && (add==amSEARCH_LOCAL || add==amUNLINK)) return NULL; /* omit global symbols */
    stackcount++;
    currsym=&(currstack->next_in_list);
    while(*currsym) {
      prelink=*currsym;
      symbolcount++;
      if ((*currsym)->type==type && !strcmp(name,(*currsym)->name)) {  /* do the types and names match ? */
	if (add==amUNLINK) {
	  unlink_from_sym(*currsym);
	} else {
	  if ((*currsym)->link) {
	    currsym=&((*currsym)->link);
	    linked=TRUE;
	  }
	  if (cmd) link_to_sym(cmd,prelink); /* associate command with symbol */
	  if (infolevel>=DIAGNOSTIC) {
	    sprintf(string,"found%ssymbol '%s' after searching %d symbol(s) in %d stack(s)",
		    linked?" linked ":" ",name,symbolcount,stackcount);
	    error(DIAGNOSTIC,string);
	  }
	  return *currsym; /* give back address */
	}
      }
      currsym=&((*currsym)->next_in_list); /* try next entry */
    }
    /* not found in first list */
    if (add==amSEARCH_VERY_LOCAL) return NULL;
    if (add==amADD_LOCAL) {
      new=create_symbol(type,name);
      if (cmd) link_to_sym(cmd,new); /* associate command with symbol */
      (*currsym)=new;
      if (infolevel>=DIAGNOSTIC) {
	sprintf(string,"created local symbol %s",name);
	error(DIAGNOSTIC,string);
      }
      return new;
    }
    currstack=currstack->prev_in_stack;
  } 
  if (add==amADD_GLOBAL) {
    new=create_symbol(type,name);
    if (cmd) link_to_sym(cmd,new);
    (*currsym)=new;
    if (infolevel>=DIAGNOSTIC) {
      sprintf(string,"created global symbol %s",name);
      error(DIAGNOSTIC,string);
    }
    return new;
  }
  return NULL;
}


void link_symbols(struct symbol *from,struct symbol *to) { /* link one symbol to the other */
  from->link=to;
  if (infolevel>=DIAGNOSTIC) {
    sprintf(string,"linking symbol '%s' to '%s'",from->name,to->name);
    error(DIAGNOSTIC,string);
  }
}


void create_retval(int is,int should) /* create command 'cRETVAL' */
{
  struct command *cmd;
  
  cmd=add_command(cRETVAL,NULL);
  cmd->args=is;
  cmd->tag=should;
}


void retval(struct command *cmd) /* check return value of function */
{
  int is,should;
  struct stackentry *s;

  is=cmd->args;
  should=cmd->tag;
  if (is==should) {
    /* okay, function returns expected type */
  } else if (is==ftNONE) { /* no element on stack, create one */
    s=push();
    if (should==ftNUMBER) {
      s->type=stNUMBER;
      s->value=0.0;
    } else {
      s->type=stSTRING;
      s->pointer=my_strdup("");
    }
  } else {
    sprintf(string,"subroutine returns %s but should return %s",
	    (is==ftSTRING)?"a string":"a number",(should==ftSTRING)?"a string":"a number");
    error(ERROR,string);
  }
  if (infolevel>=DIAGNOSTIC) {
    s=stackhead->prev;
    if (s->type==stNUMBER) 
      sprintf(string,"subroutine returns number %g",s->value);
    else if (s->type==stSTRING)
      sprintf(string,"subroutine returns string '%s'",(char *)s->pointer);
    else
      sprintf(string,"subroutine returns something strange (%d)",s->type);
    error(DIAGNOSTIC,string);
  }
  swap();
}


void create_endfunction(void) /* create command cEND_FUNCTION */
{
  struct command *cmd;

  cmd=add_command(cEND_FUNCTION,NULL);
  link_label(cmd);
}


void dump_sym(void) /* dump the stack of lists of symbols */
{
  struct symstack *currstack;
  struct symbol **currsym;
  struct command *currcmd;
  int count;
  char countstring[20];
  
  /* go through all lists */
  error(DUMP,"head of symbol stack");
  currstack=symhead;
  while(currstack) {   /* search 'til last element of stack */
    currsym=&(currstack->next_in_list);
    string[0]='\0';
    while(*currsym) {
      switch((*currsym)->type) {
      case sySTRING: strcat(string," STRING:"); break;
      case syNUMBER: strcat(string," NUMBER:"); break;
      case syFREE: strcat(string," FREE:"); break;
      case syARRAY: strcat(string," ARRAY:"); break;
      default:sprintf(string," UNKNOWN:"); break;
      }
      strcat(string,(*currsym)->name);

      /* count commands associated with this symbol */
      count=0;
      currcmd=(*currsym)->firstassoc;
      while(currcmd) {
	currcmd=currcmd->nextassoc;
	count++;
      }
      sprintf(countstring,":%d",count);
      strcat(string,countstring);

      currsym=&((*currsym)->next_in_list); /* try next entry */
    }
    error(DUMP,string);
    currstack=currstack->prev_in_stack;
  } 
  error(DUMP,"root of symbol stack");
  return;
}


static struct symbol *create_symbol(int type,char *name) /* create a new symbol */
{
  struct symbol *new;

  new=my_malloc(sizeof(struct symbol));
  new->type=type;
  new->next_in_list=NULL;
  new->name=my_strdup(name);
  new->firstassoc=NULL;
  new->pointer=NULL;
  new->args=NULL;
  new->value=0.0;
  new->link=NULL;
  
  return new;
}


static void link_to_sym(struct command *cmd,struct symbol *sym) /* link command into list associated with symbol */
{
  if (infolevel>=DIAGNOSTIC) {
    sprintf(string,"associating command '%s' with symbol '%s'",explanation[cmd->type],sym->name);
    error(DIAGNOSTIC,string);
  }
  cmd->nextassoc=sym->firstassoc;
  sym->firstassoc=cmd;
}


static void unlink_from_sym(struct symbol *sym) /* unlink all commands from symbol */
{
  struct command *currcmd;

  currcmd=sym->firstassoc;
  while(currcmd) { /* loop through list of associated commands */
    if (currcmd->pointer) {
      if (infolevel>=DIAGNOSTIC) {
	sprintf(string,"removing reference to symbol '%s' from command '%s'",
		sym->name,explanation[currcmd->type]);
	error(DIAGNOSTIC,string);
      }
      currcmd->pointer=NULL;
    }
    currcmd=currcmd->nextassoc;
  }
  sym->firstassoc=NULL;
}


void create_function_or_array(int type,char *name) /* create command 'function_or_array' */
{
  struct command *cmd;

  cmd=add_command(type,NULL);
  cmd->pointer=my_strdup(name);
}


void function_or_array(struct command *cmd) /* decide whether to perform function or array */
{
  struct command *fu;
  
  fu=search_label(cmd->pointer,smSUB|smLINK,NULL);
  if (fu) {
    cmd->type=cCALL;
    error(DIAGNOSTIC,"converting FUNCTION_OR_ARRAY to FUNCTION");
  } else {
    cmd->name=cmd->pointer;
    cmd->pointer=NULL;
    if (cmd->type==cFUNCTION_OR_ARRAY)
      cmd->tag=CALLARRAY;
    else
      cmd->tag=CALLSTRINGARRAY;
    cmd->type=cDOARRAY;
    cmd->args=-1;
    error(DIAGNOSTIC,"converting FUNCTION_OR_ARRAY to ARRAY");
  }
}


void swap() /* swap topmost elements on stack */
{
  struct stackentry *a,*b;
  
  if ((a=stackhead->prev)==NULL || (b=a->prev)==NULL) {
    error(ERROR,"Nothing to swap on stack !");
    return;
  }
  a->prev=b->prev;b->next=a->next;   /* just swap the pointers */
  a->next=b;b->prev=a;
  stackhead->prev=b;
  (a->prev)->next=a;
}


struct stackentry *push() 
/* push element on stack and enlarge it*/
{
  struct stackentry *new;
  
  if (!stackhead->next) { /* no next element */
    /* create new element */
    new=(struct stackentry *)my_malloc(sizeof(struct stackentry)); 
    /* and initialize it */
    new->next=NULL;  
    new->value=0.0;
    new->type=stFREE;
    new->prev=stackhead;
    new->pointer=NULL;
    stackhead->next=new;
  }
  stackhead=stackhead->next; /* advance head */
  /* any content is set free */
  if ((stackhead->prev)->pointer!=NULL && (stackhead->prev)->type==stSTRING) 
    free((stackhead->prev)->pointer);
  (stackhead->prev)->pointer=NULL;
  return stackhead->prev;
}


struct stackentry *pop(int etype)
/* pops element to memory and looks for pop-error */
{
  static char expected[50];
  static char found[50];
  int ftype;
  struct stackentry *s;

  /* test if there is something on the stack */
  if (stackhead==stackroot) {
    error(FATAL,"Popped too much.");
    return stackhead;
  }
  stackhead=stackhead->prev; /* move down in stack */
  ftype=stackhead->type;
  if (etype==ftype || etype==stANY || (etype==stSTRING_OR_NUMBER && (ftype==stNUMBER || ftype==stSTRING))) 
    return stackhead;  /* this is your value; use it quickly ! */
  
  /* expected and found don't match */
  stackdesc(etype,expected);
  stackdesc(ftype,found);
  sprintf(string,"expected %s but found %s",expected,found);
  if (etype==stNUMBER || etype==stSTRING || etype==stSTRING_OR_NUMBER) {
    s=push();
    if (etype==stNUMBER) {
      s->type=stNUMBER;
      s->value=0.0;
    } else {
      s->type=stSTRING;
      s->pointer=my_strdup("");
    }      
    error(ERROR,string);
    return s;
  } else {
    error(FATAL,string);
  }
  return stackhead;
}


static void stackdesc(int type,char *desc) /* give back string describing stackentry */
{
  switch(type) {
  case stGOTO: strcpy(desc,"a goto");break; 
  case stSTRING: strcpy(desc,"a string");break;
  case stSTRINGARRAYREF: strcpy(desc,"a reference to a string array");break;
  case stNUMBER: strcpy(desc,"a number");break;
  case stNUMBERARRAYREF: strcpy(desc,"a reference to a numeric array");break;
  case stLABEL: strcpy(desc,"a label");break;
  case stRETADD: strcpy(desc,"a return address for gosub");break;
  case stRETADDCALL: strcpy(desc,"a return address for a subroutine");break;
  case stFREE: strcpy(desc,"nothing");break;
  case stROOT: strcpy(desc,"the root of the stack");break;
  case stANY: strcpy(desc,"anything");break;
  case stSTRING_OR_NUMBER: strcpy(desc,"a string or a number");break;
  }
}
 
   
void pushname(char *name) /* bison: push a name on stack */
{
  struct stackentry *s;
  
  s=push();
  s->pointer=my_strdup(name);
  s->type=stSTRING;
}


void pushlabel() /* bison: generate goto and push label on stack */
{
  char *st;
  struct stackentry *en;
  
  st=(char *) my_malloc(sizeof(char)*20);
  sprintf(st,"***%d",labelcount);
  labelcount++;
  create_goto(st);
  en=push();
  en->type=stLABEL;
  en->pointer=st;
}


void poplabel() /* bison: pops a label and generates the matching command */
{
  create_label(pop(stLABEL)->pointer,cLABEL);  /* and create it */
}


void pushgoto() /* bison: generate label and push goto on stack */
{
  char *st;
  struct stackentry *en;
  
  st=(char *) my_malloc(sizeof(char)*20);
  sprintf(st,"***%d",labelcount);
  labelcount++;
  create_label(st,cLABEL);
  en=push();
  en->type=stGOTO;
  en->pointer=st;
}


void popgoto() /* bison: pops a goto and generates the matching command */
{
  create_goto(pop(stGOTO)->pointer);  /* and create it */
}


void storelabel() /* bison: push label on stack */
{
  char *st;
  struct stackentry *en;
  
  st=(char *)my_malloc(sizeof(char)*20);
  sprintf(st,"***%d",labelcount);
  labelcount++;
  en=push();
  en->type=stLABEL;
  en->pointer=st;
}


void matchgoto() /* bison: generate goto matching label on stack */
{
  create_goto(stackhead->prev->pointer);
}


void create_pushdbl(double value) /* create command 'cPUSHDBL' */
{
  struct command *cmd;
  
  cmd=add_command(cPUSHDBL,NULL);
  cmd->pointer=my_malloc(sizeof(double));
  *(double *)(cmd->pointer)=value;
}


void pushdbl(struct command *cmd) 
{
  /* push double onto stack */
  struct stackentry *p;
  
  p=push();
  p->value= *(double *)cmd->pointer;
  p->type=stNUMBER;
}


void pushfree()   /* push free entry onto stack */
{
  push()->type=stFREE;
}


void pushdblsym(struct command *cmd) 
{
  /* push double symbol onto stack */
  struct stackentry *p;
  
  p=push();
  if (!cmd->name) error(WARNING,"invalid pushdblsym");

  if (!cmd->pointer) cmd->pointer=&(get_sym(cmd->name,cmd,syNUMBER,amADD_GLOBAL)->value);
  p->value=*(double *)cmd->pointer;
  p->type=stNUMBER;
}


void popdblsym(struct command *cmd)  /* pop double from stack */
{
  double d;
  
  d=pop(stNUMBER)->value;
  if (!cmd->pointer) cmd->pointer=&(get_sym(cmd->name,cmd,syNUMBER,amADD_GLOBAL)->value);

  *(double *)(cmd->pointer)=d;
}


void create_makelocal(char *name,int type) /* create command 'cMAKELOCAL' */
{
  struct command *cmd;

  cmd=add_command(cMAKELOCAL,name);
  cmd->args=type;
}


void makelocal(struct command *cmd) /* makes symbol local */
{
  if (get_sym(cmd->name,NULL,cmd->args,amSEARCH_VERY_LOCAL)) {
    sprintf(string,"'%s' already defined within this subroutine",strip(cmd->name));
    error(ERROR,string);
    return;
  }
  get_sym(cmd->name,NULL,cmd->args,amUNLINK);
  get_sym(cmd->name,NULL,cmd->args,amADD_LOCAL);
}


void numparam(void) /* count number of function parameters */
{
  struct symbol *sym;

  get_sym(dotify("numparams"),NULL,syNUMBER,amUNLINK);
  sym=get_sym(dotify("numparams"),NULL,syNUMBER,amADD_LOCAL);
  sym->value=abs(count_args(FALSE));
}


void create_makestatic(char *name,int type) /* create command 'cMAKESTATIC' */
{
  struct command *cmd;

  cmd=add_command(cMAKESTATIC,name);
  cmd->pointer=current_function;
  cmd->args=type;
}


void makestatic(struct command *cmd) /* makes symbol static */
{
  struct symbol *l,*g;
  static char global[200];
  
  if (get_sym(cmd->name,NULL,cmd->args,amSEARCH_VERY_LOCAL)) {
    sprintf(string,"'%s' already defined within this subroutine",strip(cmd->name));
    error(ERROR,string);
    return;
  }
  /* construct a unique name as library.variable_name@function_name */
  strcpy(global,cmd->name);
  strcat(global,"@");
  strcat(global,cmd->pointer);
  /* create global variable with unique name */
  g=get_sym(global,NULL,cmd->args,amADD_GLOBAL);
  /* create local variable */
  l=get_sym(cmd->name,NULL,cmd->args,amADD_LOCAL);
  /* link those two together */
  link_symbols(l,g);
}


void create_arraylink(char *name,int type) /* create command 'cARRAYLINK' */
{
  struct command *cmd;

  cmd=add_command(cARRAYLINK,name);
  cmd->pointer=current_function;
  cmd->args=type;
}


void arraylink(struct command *cmd) /* link a local symbol to a global array */
{
  struct symbol *l,*g;
  struct array *ar;
  
  if (get_sym(cmd->name,NULL,cmd->args,amSEARCH_VERY_LOCAL)) {
    sprintf(string,"'%s()' already defined within this subroutine",strip(cmd->name));
    error(ERROR,string);
    return;
  }
  /* get globally defined array */
  g=get_sym(pop(cmd->args)->pointer,NULL,syARRAY,amSEARCH);
  /* create local array */
  l=get_sym(cmd->name,cmd,syARRAY,amADD_LOCAL);
  if (!l) return;
  if (!g || !g->pointer) { /* no global array supplied, create one */
    error(DIAGNOSTIC,"creating dummy array");
    ar=create_array((cmd->args==stNUMBERARRAYREF)?'d':'s',0);
    l->pointer=ar;
    if (infolevel>=DIAGNOSTIC) {
      sprintf(string,"creating 0-dimensional dummy array '%s()'",cmd->name);
      error(DIAGNOSTIC,string);
    }
  } else {
    /* link those two together */
    link_symbols(l,g);
  }
}


void create_pusharrayref(char *name,int type) /* create command 'cPUSHARRAYREF' */
{
  struct command *cmd;

  cmd=add_command(cPUSHARRAYREF,name);
  cmd->args=type;
}


void pusharrayref(struct command *cmd) /* push an array reference onto stack */
{
  struct stackentry *s;
  s=push();
  s->type=cmd->args;
  s->pointer=my_strdup(cmd->name);
}


void create_require(int type) /* create command 'cREQUIRE' */
{
  struct command *cmd;
  
  cmd=add_command(cREQUIRE,NULL);
  cmd->args=type;
}


void require(struct command *cmd) /* check, that item on stack has right type */
{
  char *expected,*supplied;
  struct stackentry *s;

  if (stackhead->prev->type==cmd->args) return; /* okay, they match */
      
  if (stackhead->prev->type==stFREE) { /* no argument supplied, create it */
    s=push();
    if (cmd->args==stSTRING) {
      s->type=stSTRING;
      s->pointer=my_strdup("");
      return;
    } else if (cmd->args==stNUMBER) {
      s->type=stNUMBER;
      s->value=0.0;
      return;
    } else {
      /* create array */
      s->type=cmd->args;
      s->pointer=NULL;
      return;
    }
  }
    
  s=stackhead->prev;
  if (s->type==stSTRING) 
    supplied="string";
  else if (s->type==stNUMBER) 
    supplied="number";
  else if (s->type==stSTRINGARRAYREF)
    supplied="string array";
  else if (s->type==stNUMBERARRAYREF)
    supplied="numeric array";
  else if (s->type==stFREE) 
    supplied="nothing";
  else 
    supplied="something strange";

  if (cmd->args==stSTRING) 
    expected="string";
  else if (cmd->args==stNUMBER) 
    expected="number";
  else if (cmd->args==stSTRINGARRAYREF)
    expected="string array";
  else if (cmd->args==stNUMBERARRAYREF)
    expected="numeric array";
  else if (cmd->args==stFREE)
    expected="nothing";
  else
    expected="something strange";

  sprintf(string,"invalid subroutine call: %s expected, %s supplied",expected,supplied);
  error(ERROR,string);
}


void create_dblbin(char c) /* create command for binary double operation */
{
  switch(c) {
  case '+':add_command(cDBLADD,NULL);break;
  case '-':add_command(cDBLMIN,NULL);break;
  case '*':add_command(cDBLMUL,NULL);break;
  case '/':add_command(cDBLDIV,NULL);break;
  case '^':add_command(cDBLPOW,NULL);break;
  }
  /* no specific information needed */
}


void dblbin(struct command *cmd) /* compute with two numbers from stack */
{
  struct stackentry *d;
  double a,b,c;
  
  b=pop(stNUMBER)->value;
  a=pop(stNUMBER)->value;
  d=push();
  switch(cmd->type) {
  case(cDBLADD):c=a+b; break;
  case(cDBLMIN):c=a-b; break;
  case(cDBLMUL):c=a*b; break;
  case(cDBLDIV): 
    if (fabs(b)<DBL_MIN) {
      sprintf(string,"Division by zero, set to %g",DBL_MAX);
      error(NOTE,string);
      c=DBL_MAX;}
    else
      c=a/b;
    break;
  case(cDBLPOW):
    if (b==2) 
      c=a*a;
    else if (a<0) {
      error(ERROR,"Power of negative value. Don't now what to do");
      return;}
    else
      c=exp(b*log(a));
    break;
  }
  d->value=c;
  d->type=stNUMBER;
}


void negate() /* negates top of stack */
{
  stackhead->prev->value=-stackhead->prev->value;
}


void pushstrptr(struct command *cmd)  /* push string-pointer onto stack */
{
  struct stackentry *p;
  
  p=push();
  if (!cmd->pointer) cmd->pointer=&(get_sym(cmd->name,cmd,sySTRING,amADD_GLOBAL)->pointer);
  p->pointer=*(char **)cmd->pointer;
  p->type=stSTRING;
}


void pushstrsym(struct command *cmd)  /* push string-symbol onto stack */
{
  struct stackentry *p;
  
  p=push();
  if (!cmd->pointer) cmd->pointer=&(get_sym(cmd->name,cmd,sySTRING,amADD_GLOBAL)->pointer);
  p->pointer=my_strdup(*(char **)cmd->pointer);
  p->type=stSTRING;
}


void popstrsym(struct command *cmd) /* pop string from stack */
{
  if (!cmd->name) return;
  if (!cmd->pointer) cmd->pointer= &(get_sym(cmd->name,cmd,sySTRING,amADD_GLOBAL)->pointer);
  if (*(char **)cmd->pointer!=NULL) free(*(char **)cmd->pointer);
  *(char **)cmd->pointer=my_strdup(pop(stSTRING)->pointer);
}


void create_pushstr(char *s) /* creates command pushstr */
{
  struct command *cmd;
  
  cmd=add_command(cPUSHSTR,NULL);
  cmd->pointer=my_strdup(s); /* store string */
}


void pushstr(struct command *cmd) 
{
  /* push string onto stack */
  struct stackentry *p;
  
  p=push();
  p->pointer=my_strdup((char *)cmd->pointer);
  p->type=stSTRING;
}


void duplicate(void) /* duplicate topmost element of stack */
{
  struct stackentry *s;
  double actual;
  
  actual=stackhead->prev->value;
  s=push();
  s->type=stNUMBER;
  s->value=actual;
}


void create_goto(char *label) /* creates command goto */
{
  struct command *cmd;
  
  cmd=add_command(cGOTO,NULL);
  /* specific info */
  cmd->pointer=my_strdup(label);
}


void create_gosub(char *label) /* creates command gosub */
{
  struct command *cmd;
  
  cmd=add_command(cGOSUB,NULL);
  /* specific info */
  cmd->pointer=my_strdup(label);
}


void create_call(char *label) /* creates command function call */
{
  struct command *cmd;
  
  cmd=add_command(cCALL,NULL);
  /* specific info */
  cmd->pointer=my_strdup(label);
}


static void link_label(struct command *cmd) /* link label into list of labels */
{
  if (!labelroot) 
    labelroot=cmd;
  else
    labelhead->nextassoc=cmd;
  labelhead=cmd;
}


static struct command *search_label(char *name,int type,int *fu) /* search label */
{
  struct command *curr;
  int look=FALSE;

  if (fu) *fu=0;
  curr=labelroot;
  while(curr) {
    if ((type&smSUB) && curr->type==cUSER_FUNCTION && !strcmp(curr->pointer,name)) return curr;
    if ((type&smLINK) && curr->type==cSUBLINK && !strcmp(curr->pointer,name)) return curr->next;
    if ((type&smLABEL) && curr->type==cLABEL && !strcmp(curr->pointer,name)) return curr;
    if (fu && *fu>=0 && curr->cnt>=current->cnt) look=TRUE;
    if (fu && *fu<0) look=FALSE;
    if (look && curr->type==cUSER_FUNCTION) (*fu)++;
    if (look && curr->type==cEND_FUNCTION) (*fu)--;
    curr=curr->nextassoc;
  }
  return NULL;
}


void jump(struct command *cmd) 
/* jump to specific Label; used as goto, gosub or function call */
{
  struct command *label;
  struct stackentry *ret;
  int type;
  int functions;
  
  type=cmd->type;

  type=cmd->type;
  if (type==cGOSUB || type==cQGOSUB || type==cCALL || type==cQCALL) {
    /* leave return address for return */
    ret=push();
    ret->pointer=current;
    if (type==cGOSUB || type==cQGOSUB) {
      ret->type=stRETADD;
    } else {
      ret->type=stRETADDCALL;
      reshufflestack(ret);
    }
  }

  if (type==cQGOSUB || type==cQGOTO || type==cQCALL) {
    current=(struct command *)cmd->pointer; /* use remembered address */
    return;
  }
  label=search_label(cmd->pointer,smSUB|smLINK|smLABEL,&functions);
  if (label) {
      /* found right label */
    current=label; /* jump to new location */
    /* use the address instead of the name next time */
    cmd->pointer=label;
    switch(cmd->type) {
    case cGOTO: cmd->type=cQGOTO; break;
    case cGOSUB: cmd->type=cQGOSUB; break;
    case cCALL: cmd->type=cQCALL; break;
    }
    if (type!=cCALL && functions) error(ERROR,"can not jump into or out of subroutine");
  } else {
    /* label not found */
    sprintf(string,"Can't find %s '%s'",(type==cCALL)?"subroutine":"label",strip((char *)cmd->pointer));
    error(ERROR,string);
  }
}


static void reshufflestack(struct stackentry *ret) /* reorganize stack for function call */
{
  struct stackentry *a,*b,*c;
  struct stackentry *top,*bot;
  struct stackentry *ttop,*bbot;
  int args;


  /* this is a function call; revert stack and shuffle return address to bottom */
  /* push address below parameters */
  args=0;
  top=a=ret->prev;
  while(a->type!=stFREE) {
    a=a->prev;
    args++;
  }
  bot=a->next;
  b=a->prev;
  /* remove ret */
  ret->prev->next=ret->next;
  ret->next->prev=ret->prev;
  /* squeeze ret between a and b */
  ret->next=a;
  a->prev=ret;
  b->next=ret;
  ret->prev=b;
  /* revert stack between top and bot */
  if (args>1) {
    a=bot;
    b=a->next;
    bbot=bot->prev;
    ttop=top->next;
    for(;args>1;args--) {
      a->prev=b;
      c=b->next;
      b->next=a;
      a=b;
      b=c;
    }
    bot->next=ttop;
    bot->next->prev=bot;
    top->prev=bbot;
    top->prev->next=top;
  }
}


void myreturn(struct command *cmd) /* return from gosub of function call */
{
  struct stackentry *address;

  address=pop(stANY);
  if (cmd->type==cRET_FROM_FUN) {
    if (address->type!=stRETADDCALL) {
      error(ERROR,"RETURN from a subroutine without CALL");
      return;
    }
  } else {
    if (address->type!=stRETADD) {
      error(ERROR,"RETURN without GOSUB");
      return;
    }
  }
  current=(struct command *)address->pointer;
  return;
}


void create_label(char *label,int type) /* creates command label */
{
  struct command *cmd;
  
  /* check, if label is duplicate */
  if (search_label(label,smSUB|smLINK|smLABEL,NULL)) {
    sprintf(string,"duplicate %s '%s'",(type==cLABEL)?"label":"subroutine",strip(label));
    error(ERROR,string);
    return;
  }
  
  cmd=add_command(type,NULL);
  /* store label */
  cmd->pointer=my_strdup(label);
  link_label(cmd);
}


void create_sublink(char *label) /* create link to subroutine */
{
  char global[200];
  char *dot;
  struct command *cmd;

  if (!in_libfile) return;
  dot=strchr(label,'.');
  strcpy(global,"main");
  strcat(global,dot);
  
  /* check, if label is duplicate */
  if (search_label(global,smSUB|smLINK|smLABEL,NULL)) {
    sprintf(string,"duplicate subroutine '%s'",strip(global));
    error(ERROR,string);
    return;
  }

  cmd=add_command(cSUBLINK,NULL);
  /* store label */
  cmd->pointer=my_strdup(global);
  link_label(cmd);
}


void decide() /*  skips next command, if not 0 on stack */
{
  if (pop(stNUMBER)->value!=0) current=current->next; /* skip one command */
}


void create_dim(char *name,char type) /* create command 'dim' */
/* type can be 's'=string or 'd'=double Array */
{ 
  struct command *cmd;
  
  cmd=add_command(cDIM,name);
  cmd->tag=type; /* type: string or double */
  cmd->args=-1;
}


void dim(struct command *cmd) /* get room for array */
{
  struct array *nar,*oar;
  char *nul;
  int ntotal,ototal,esize,i,j;
  int ind[10],nbounds[10],larger;
  struct symbol *s;

  if (cmd->args<0) cmd->args=count_args(FALSE);
  if (cmd->args<0) {
    error(ERROR,"only numerical indices allowed for arrays");
    return;
  }
  s=get_sym(cmd->name,cmd,syARRAY,amADD_LOCAL);
  if (search_label(cmd->name,smSUB|smLINK,NULL)) {
    sprintf(string,"array '%s()' conflicts with user subroutine",strip(cmd->name));
    error(ERROR,string);
    return;
  }

  /* check for dimensions */
  if (cmd->args>10) {
    error(ERROR,"more than 10 indices");
    return;
  }
  oar=s->pointer;
  if (oar) {
    /* check, if old and new array are compatible */
    if (cmd->args!=oar->dimension) {
      sprintf(string,"cannot change dimension of '%s()' from %d to %d",
	      strip(cmd->name),oar->dimension,cmd->args);
      error(ERROR,string);
    }
  }
  /* check, if redim is actually needed */
  for(i=0;i<10;i++) nbounds[i]=0;
  larger=FALSE;
  for(i=0;i<cmd->args;i++) {
    nbounds[i]=1+(int)pop(stNUMBER)->value;
    if (nbounds[i]<=1) {
      sprintf(string,"%d. index is less or equal zero",i+1);
      error(ERROR,string);
      return;
    }
    if (oar) {
      if (nbounds[i]>oar->bounds[i]) 
	larger=TRUE;
      else
	nbounds[i]=oar->bounds[i];
    }
  }
  pop(stFREE); /* remove left over stFREE */
  if (oar && !larger) return; /* new array won't be larger than old one */

  /* create array */
  nar=create_array(cmd->tag,cmd->args);

  /* count needed memory */
  ntotal=1;
  for(i=0;i<nar->dimension;i++) {
    (nar->bounds)[i]=nbounds[i];
    ntotal*=nbounds[i];
  }
  esize=(nar->type=='s')?sizeof(char *):sizeof(double); /* size of one array element */
  nar->pointer=my_malloc(ntotal*esize);

  if (oar) { /* array already exists */
    /* unlink array from associated commands */
    unlink_from_sym(s);
    ototal=1;
    for(i=0;i<oar->dimension;i++) ototal*=(oar->bounds)[i];
  }

  /* initialize Array */
  for(i=0;i<ntotal;i++) {
    if (nar->type=='s') { 
      nul=my_malloc(sizeof(char));
      *nul='\0';
      ((char **)nar->pointer)[i]=nul;
    } else {
      ((double *)nar->pointer)[i]=0.0;
    }
  }  
  
  if (oar) { /* copy contents of old array onto new */
    for(i=0;i<ototal;i++) {
      off_to_ind(i,oar->bounds,ind);
      j=ind_to_off(ind,nar->bounds);
      if (nar->type=='s') {
	free(((char **)nar->pointer)[j]);
	((char **)nar->pointer)[j]= ((char **)oar->pointer)[i];
      } else {
	((double *)nar->pointer)[j]= ((double *)oar->pointer)[i];
      }
    }
    free(oar->pointer);
    free(oar);
  }
  
  s->pointer=nar;
  cmd->pointer=nar;
}


static int ind_to_off(int *ind,int *bound) /* convert array of indices to single offset */
{
  int i;
  int cur,off;

  off=0;
  cur=1;
  for(i=0;i<10 && bound[i];i++) {
    off+=ind[i]*cur;
    cur*=bound[i];
  }
  return off;
}


static void off_to_ind(int off,int *bound,int *ind) /* convert a single offset to an array of indices */
{
  int i;
  int cur;
  
  cur=1;
  for(i=0;i<10;i++) {
    if (bound[i]) cur*=bound[i];
    ind[i]=0;
  }
  for(i=9;i>=0;i--) {
    if (bound[i]) {
      cur/=bound[i];
      ind[i]=off/cur;
      off-=ind[i]*cur;
    } else {
      ind[i]=0;
    }
  }
}


void query_array(struct command *cmd) /* query array */
{
  int index;
  struct stackentry *s;
  struct array *ar;
  struct symbol *sym;

  if (cmd->type==cARSIZE) index=(int)pop(stNUMBER)->value;

  s=pop(stANY);
  if (s->type!=stSTRINGARRAYREF && s->type!=stNUMBERARRAYREF) {
    error(ERROR,"expecting a reference to an array");
    return;
  }
  
  if (!cmd->pointer) {
    sym=get_sym(s->pointer,cmd,syARRAY,amSEARCH);
    if (!sym || !sym->pointer) {
      sprintf(string,"array '%s()' is not defined",strip(s->pointer));
      error(ERROR,string);
      return;
    }
    cmd->pointer=sym->pointer;
  }

  ar=cmd->pointer;
  if (cmd->type==cARSIZE && (index<1 || index>ar->dimension)) {
    sprintf(string,"only indices between 1 and %d allowed",ar->dimension);
    error(ERROR,string);
    return;
  }
  s=push();
  s->type=stNUMBER;
  if (cmd->type==cARSIZE)
    s->value=ar->bounds[ar->dimension-index]-1;
  else
    s->value=ar->dimension;
  
  return;
}


void create_doarray(char *symbol,int command) /* creates array-commands */ 
{
  struct command *cmd;

  cmd=add_command(cDOARRAY,symbol);
  cmd->tag=command; /* operation to perform */
  cmd->args=-1;
}


void doarray(struct command *cmd) /* call an array */
{
  struct array *ar;
  struct stackentry *stack;
  struct symbol *sym;
  void *p;
  char **str;
  double *dbl;
  int i,j,bnd,index,cur,rval;
 

  if (!cmd->pointer) {
    sym=get_sym(cmd->name,cmd,syARRAY,amSEARCH);
    if (!sym || !sym->pointer) {
      sprintf(string,"'%s()' is neither array nor subroutine",strip(cmd->name));
      error(ERROR,string);
      return;
    }
    cmd->pointer=sym->pointer;
  }
  rval=(current->tag==CALLARRAY || current->tag==CALLSTRINGARRAY || current->tag==GETSTRINGPOINTER);
  if (cmd->args<0) cmd->args=count_args(!rval);
  if (cmd->args<0) {
    error(ERROR,"only numerical indices allowed for arrays");
    return;
  }
  cmd->args=abs(cmd->args);
  if (!cmd->args) { /* no indizes supplied, create a reference to an array */
    pop(stFREE); /* remove left over stFREE */
    stack=push();
    if (cmd->tag==CALLARRAY)
      stack->type=stNUMBERARRAYREF;
    else
      stack->type=stSTRINGARRAYREF;
    stack->pointer=my_strdup(cmd->name);
    return;
  }

  ar=cmd->pointer;

  if (!ar->dimension) {
    sprintf(string,"array parameter '%s()' has not been supplied",strip(cmd->name));
    error(ERROR,string);
    return;
  }
  if (cmd->args!=ar->dimension) {
    sprintf(string,"%d indices supplied, %d expected for '%s()'",cmd->args,ar->dimension,strip(cmd->name));
    error(ERROR,string);
    return;
  }

  if (!rval) stack=pop(stSTRING_OR_NUMBER);
  index=0;
  cur=1;
  for(i=0;i<ar->dimension;i++) {
    bnd=(ar->bounds[i]);
    j=(int)pop(stNUMBER)->value;
    if (j<0 || j>=bnd) {
      sprintf(string,"index %d (=%d) out of range",ar->dimension-i,j);
      error(ERROR,string);
      return;
    }
    index+=j*cur;
    cur*=bnd;
  }
  
  pop(stFREE); /* remove left over stFREE */
  if (rval) stack=push();
  
  p=ar->pointer;
  switch(current->tag) {
  case CALLARRAY:
    dbl=(double *)p+index;
    stack->value= *dbl;
    stack->type=stNUMBER;
    break;
  case ASSIGNARRAY:
    dbl=(double *)p+index;
    *dbl=stack->value;
    break;
  case CALLSTRINGARRAY:
    str=((char **)p+index);
    stack->pointer=my_strdup(*str);
    stack->type=stSTRING;
    break;
  case ASSIGNSTRINGARRAY:
    str=((char **)p+index);
    if (*str!=NULL) free(*str);
    *str=my_strdup(stack->pointer);
    break;
  case GETSTRINGPOINTER:
    str=((char **)p+index);
    stack->pointer=*str;
    stack->type=stSTRING;
    break;
  }
}


struct array *create_array(int type,int dimension) /* create an array */
{
  struct array *ar;
  int i;

  ar=my_malloc(sizeof(struct array));
  ar->type=(type);
  ar->dimension=dimension;
  ar->pointer=NULL;
  for(i=0;i<10;i++) ar->bounds[i]=0;

  return ar;
}


static int count_args(int skipfirst) /* count number of numeric arguments on stack */
{
  int i=0;
  int sign=1;
  struct stackentry *curr;

  curr=stackhead->prev;
  if (skipfirst) curr=curr->prev;
  while(curr) {
    if (curr->type==stFREE) return i*sign;
    if (curr->type!=stNUMBER) sign=-1;
    curr=curr->prev;
    i++;
  }
  return -1;
}


void skipper()
/* used for on_goto/gosub, skip specified number of commands */
{
  int i,len;
  struct command *ahead; /* command to follow */
  
  len=(int)pop(stNUMBER)->value;
  i=1;
  current=current->next; /* advance to first goto/gosub */
  for(i=1;i<len;i++) {
    ahead=current->next->next; /* skip interleaving findnop statement */
    if (ahead->type==cNOP) 
      break;
    else
      current=ahead;
  }
}


void skiponce(struct command *cmd) /* skip next command exectly once */
{
  if (cmd->tag) current=current->next;
  cmd->tag=0;
}


void resetskiponce(struct command *cmd) /* find and reset next skip */
{
  struct command *c;
  
  c=cmd;
  while(c->type!=cSKIPONCE) c=c->next;
  c->tag=1;
}


void findnop()
/* used for on_gosub, find trailing nop command */
{
  while(current->type!=cNOP) {
    current=current->next; /* next label */
  }
}


void forcheck(void) /* check, if for-loop is done */
{
  double start,bound,step,val;
  
  val=pop(stNUMBER)->value;
  step=pop(stNUMBER)->value;
  bound=pop(stNUMBER)->value;
  start=stackhead->prev->value;
  if ((val<=bound && val>=start) || (val<=start && val>=bound)) 
    stackhead->prev->value=1.;
  else
    stackhead->prev->value=0.;
}


void forincrement(void) /* increment value on stack */
{
/* expecting on stack: BOUND,STEP,VAL,stackhead
  where for VAL=START to BOUND step STEP */
  stackhead->prev->value+=stackhead->prev->prev->value;
}


void startfor(void) /* compute initial value of for-variable */
{
  struct stackentry *p;
  
  p=push();
  p->value=stackhead->prev->prev->prev->prev->value-stackhead->prev->prev->value;
  p->type=stNUMBER;
  
  return;
}


